# Standard Library
import csv

# Third Party
from linkml_runtime.dumpers import YAMLDumper

# Local
from ai_atlas_nexus.ai_risk_ontology import (
    Container,
    Documentation,
    Risk,
    RiskGroup,
    RiskTaxonomy,
)


def create_container_object() -> Container:
    # Create risk taxonomy paper
    documents = [
        Documentation(
            **{
                "id": "arxiv.org/2408.12622",
                "name": "The AI Risk Repository: A Comprehensive Meta-Review, Database, and Taxonomy of Risks From Artificial Intelligence",
                "description": "The risks posed by Artificial Intelligence (AI) are of considerable concern to academics, auditors, policymakers, AI companies, and the public. However, a lack of shared understanding of AI risks can impede our ability to comprehensively discuss, research, and react to them. This paper addresses this gap by creating an AI Risk Repository to serve as a common frame of reference. This comprises a living database of 777 risks extracted from 43 taxonomies, which can be filtered based on two overarching taxonomies and easily accessed, modified, and updated via our website and online spreadsheets. We construct our Repository with a systematic review of taxonomies and other structured classifications of AI risk followed by an expert consultation. We develop our taxonomies of AI risk using a best-fit framework synthesis. Our high-level Causal Taxonomy of AI Risks classifies each risk by its causal factors (1) Entity: Human, AI; (2) Intentionality: Intentional, Unintentional; and (3) Timing: Pre-deployment; Post-deployment. Our mid-level Domain Taxonomy of AI Risks classifies risks into seven AI risk domains: (1) Discrimination & toxicity, (2) Privacy & security, (3) Misinformation, (4) Malicious actors & misuse, (5) Human-computer interaction, (6) Socioeconomic & environmental, and (7) AI system safety, failures, & limitations. These are further divided into 23 subdomains. The AI Risk Repository is, to our knowledge, the first attempt to rigorously curate, analyze, and extract AI risk frameworks into a publicly accessible, comprehensive, extensible, and categorized risk database. This creates a foundation for a more coordinated, coherent, and complete approach to defining, auditing, and managing the risks posed by AI systems.",
                "url": "https://arxiv.org/abs/2408.12622",
                "dateCreated": "2024-08-14",
                "dateModified": "2024-08-14",
            }
        )
    ]

    # Create risk taxonomy
    taxonomies = [
        RiskTaxonomy(
            **{
                "id": "mit-ai-risk-repository",
                "name": "The AI Risk Repository",
                "description": "A comprehensive living database of over 700 AI risks categorized by their cause and risk domain.",
                "url": "https://airisk.mit.edu/",
                "version": "1",
                "dateCreated": "2024-08-16",
                "hasDocumentation": ["arxiv.org/2408.12622"],
            }
        )
    ]

    # Create risk groups
    risks = get_risks()
    risk_group_names = sorted(
        list(set([(risk["Major"], risk["Domain"]) for risk in risks]))
    )
    riskgroups = [
        RiskGroup(
            **{
                "id": "mit-ai-risk-domain-" + risk_group[0],
                "name": risk_group[1],
                "isDefinedByTaxonomy": "mit-ai-risk-repository",
            }
        )
        for risk_group in risk_group_names
    ]

    # Create risks
    risk_objects = [
        Risk(
            **{
                "id": "mit-ai-risk-subdomain-" + risk["Minor"],
                "name": risk["Sub-domain"],
                "description": risk["Description"],
                "isPartOf": "mit-ai-risk-domain-" + risk["Major"],
                "isDefinedByTaxonomy": "mit-ai-risk-repository",
            }
        )
        for risk in risks
    ]

    # Create container
    container = Container(
        documents=documents,
        taxonomies=taxonomies,
        riskgroups=riskgroups,
        risks=risk_objects,
    )
    return container


def get_risks() -> list[dict]:
    with open("resources/TheAIRiskRepositoryV3_26_03_2025.csv") as csvfile:
        importer = csv.DictReader(csvfile)
        return [row for row in importer]


with open(
    "src/ai_atlas_nexus/data/knowledge_graph/mit_ai_risk_repository_data_4_july_2025.yaml",
    "+tw",
    encoding="utf-8",
) as output_file:
    container = create_container_object()
    print(YAMLDumper().dumps(container), file=output_file)
    output_file.close()



from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Risk
from ai_atlas_nexus.blocks.risk_explorer.base import ExplorerBase
from ai_atlas_nexus.data.knowledge_graph import *


class RiskExplorer(ExplorerBase):

    def __init__(self, ontology):

        # load the data into the graph
        self._ontology = ontology
        self._risks = ontology.risks or []
        self._riskcontrols = ontology.riskcontrols or []
        self._riskincidents = ontology.riskincidents or []
        self._actions = ontology.actions or []
        self._taxonomies = ontology.taxonomies or []
        self._evaluations = ontology.evaluations or []
        self._aievaluationresults = ontology.aievalresults or []
        self._benchmarkmetadatacards = ontology.benchmarkmetadatacards or []
        self._documents = ontology.documents or []
        self._datasets = ontology.datasets or []
        self._stakeholders = ontology.stakeholders or []
        self._llmintrinsics = ontology.llmintrinsics or []
        self._adapters = ontology.adapters or []
        self._llmquestionpolicies = ontology.llmquestionpolicies or []
        self._rules = ontology.rules or []
        self._principles = ontology.principles or []

    def get_all_risks(self, taxonomy=None):
        """Get all risk definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Risk]
                Result containing a list of AI risks
        """
        risk_instances = self._risks
        if taxonomy is not None:
            risk_instances = list(
                filter(
                    lambda risk: risk.isDefinedByTaxonomy == taxonomy, risk_instances
                )
            )

        return risk_instances

    def get_risk(self, tag=None, id=None, name=None, taxonomy=None):
        """Get a risk definition from the LinkML by tag, id or name

        Args:
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            Risk
                An AI risk result
        """
        matching_risks = self._risks
        if tag is not None:
            matching_risks = list(filter(lambda risk: risk.tag == tag, matching_risks))
        if id is not None:
            matching_risks = list(filter(lambda risk: risk.id == id, matching_risks))
        if name is not None:
            matching_risks = list(
                filter(lambda risk: risk.name == name, matching_risks)
            )
        if taxonomy is not None:
            matching_risks = list(
                filter(
                    lambda risk: risk.isDefinedByTaxonomy == taxonomy, matching_risks
                )
            )

        if len(matching_risks) > 0:
            return matching_risks[0]
        else:
            print("No matching risk found")
            return None

    def _combine_related_risks(self, initial_risk):
        """Convenience function to combine the related risks from a Risk

        Args:
            initial_risk: Risk
                The risk to parse for related risks

        Returns:
            list[str]
                Result containing a list of AI risk IDs
        """
        related_risks = []
        if initial_risk.broadMatch is not None:
            related_risks.append(initial_risk.broadMatch)
        if initial_risk.closeMatch is not None:
            related_risks.append(initial_risk.closeMatch)
        if initial_risk.exactMatch is not None:
            related_risks.append(initial_risk.exactMatch)
        if initial_risk.narrowMatch is not None:
            related_risks.append(initial_risk.narrowMatch)
        if initial_risk.relatedMatch is not None:
            related_risks.append(initial_risk.relatedMatch)

        related_risk_ids = [j for i in related_risks for j in i]
        related_risks = list(
            filter(lambda risk: risk.id in related_risk_ids, self._risks)
        )
        return related_risks

    def get_related_risks(self, risk=None, tag=None, id=None, name=None, taxonomy=None):
        """Get all related risk definitions from the LinkML by the IBM risk atlas tag

        Args:
            risk: (Optional) Risk
                The Risk object to find related risks for
            id: (Optional) str
                The string ID identifying the risk
            tag: (Optional) str
                The string tag identifying the risk
            name: (Optional) str
                The string name identifying the risk
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[str]
                Result containing a list of AI risks IDs
        """
        matching_risks = self._risks

        if risk is not None:
            matching_risks = [risk]
        if tag is not None:
            matching_risks = list(filter(lambda risk: risk.tag == tag, matching_risks))
        if id is not None:
            matching_risks = list(filter(lambda risk: risk.id == id, matching_risks))
        if name is not None:
            matching_risks = list(
                filter(lambda risk: risk.name == name, matching_risks)
            )

        if len(matching_risks) > 0:
            initial_risk: Risk = matching_risks[0]
            related_risks = self._combine_related_risks(initial_risk)

            if taxonomy is not None:
                related_risks = list(
                    filter(
                        lambda risk: risk.isDefinedByTaxonomy == taxonomy, related_risks
                    )
                )

            return related_risks
        else:
            print("No matching risk found")
            return None

    def get_related_actions(
        self, risk=None, tag=None, id=None, name=None, taxonomy=None
    ):
        """Get actions for a risk from the LinkML

        Args:
            risk: (Optional) Risk
                The Risk object to find related actions for
            id: (Optional) str
                The string ID identifying the risk to find related actions for
            tag: (Optional) str
                The string tag identifying the risk to find related actions for
            name: (Optional) str
                The string name identifying the risk to find related actions for
            taxonomy: str
                (Optional) Only return actions which come from this taxonomy

        Returns:
            list[Action]
                Result containing a list of the actions which are marked as related to the specified AI risk
        """
        matching_risks = self._risks

        if risk is not None:
            matching_risks = [risk]
        if tag is not None:
            matching_risks = list(filter(lambda risk: risk.tag == tag, matching_risks))
        if id is not None:
            matching_risks = list(filter(lambda risk: risk.id == id, matching_risks))
        if name is not None:
            matching_risks = list(
                filter(lambda risk: risk.name == name, matching_risks)
            )

        if len(matching_risks) > 0:
            risk: Risk = matching_risks[0]
            actions = []

            if risk.hasRelatedAction is not None:
                actions.append(risk.hasRelatedAction)

            actions = [j for i in actions for j in i]

            if taxonomy is not None:
                actions = list(
                    filter(
                        lambda action: action.isDefinedByTaxonomy == taxonomy, actions
                    )
                )

            return actions
        else:
            print("No matching actions found")
            return None

    def get_all_actions(self, taxonomy=None):
        """Get all actions definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Action]
                Result containing a list of AI actions
        """
        action_instances = self._actions

        if taxonomy is not None:
            action_instances = list(
                filter(
                    lambda action: action.isDefinedByTaxonomy == taxonomy,
                    action_instances,
                )
            )

        return action_instances

    def get_action_by_id(self, id):
        """Get action definition from the LinkML by ID

        Args:
            id: str
                The string id for an action

        Returns:
            Action
                Result containing an AI actions
        """
        matching_actions = list(filter(lambda action: action.id == id, self._actions))

        if len(matching_actions) > 0:
            return matching_actions[0]
        else:
            print("No matching risk found")
            return None

    def get_all_taxonomies(self):
        """Get all taxonomy definitions from the LinkML

        Returns:
            list[RiskTaxonomy]
                Result containing a list of risk taxonomies
        """
        return self._taxonomies

    def get_taxonomy_by_id(self, id):
        """Get taxonomy definition from the LinkML by ID

        Args:
            id: str
                The string id for a taxonomy

        Returns:
            RiskTaxonomy
                Result containing an AI RiskTaxonomy
        """
        matching_taxonomies = list(
            filter(lambda taxonomy: taxonomy.id == id, self._taxonomies)
        )

        if len(matching_taxonomies) > 0:
            return matching_taxonomies[0]
        else:
            print("No matching taxonomy found")
            return None

    def get_related_risk_controls(
        self, risk=None, tag=None, id=None, name=None, taxonomy=None
    ):
        """Get related risk controls for a risk from the LinkML

        Args:
            risk: (Optional) Risk
                The Risk object to find related actions for
            id: (Optional) str
                The string ID identifying the risk to find related actions for
            tag: (Optional) str
                The string tag identifying the risk to find related actions for
            name: (Optional) str
                The string name identifying the risk to find related actions for
            taxonomy: str
                (Optional) The string label for a taxonomy, to filter action results by

        Returns:
            list[RiskControl]
                Result containing a list of the risk controls which are marked as related to the specified AI risk
        """
        matching_risks = self._risks

        if risk is not None:
            matching_risks = [risk]
        if tag is not None:
            matching_risks = list(filter(lambda risk: risk.tag == tag, matching_risks))
        if id is not None:
            matching_risks = list(filter(lambda risk: risk.id == id, matching_risks))
        if name is not None:
            matching_risks = list(
                filter(lambda risk: risk.name == name, matching_risks)
            )

        if len(matching_risks) > 0:
            risk: Risk = matching_risks[0]
            risk_controls = []

            if risk.isDetectedBy is not None:
                risk_controls.append(risk.isDetectedBy)

            risk_controls = [j for i in risk_controls for j in i]

            if taxonomy is not None:
                risk_controls = list(
                    filter(
                        lambda risk_control: risk_control.isDefinedByTaxonomy
                        == taxonomy,
                        risk_controls,
                    )
                )
            related_risk_controls = list(
                filter(
                    lambda risk_control: risk_control.id in risk_controls,
                    self._riskcontrols,
                )
            )
            return related_risk_controls
        else:
            print("No matching risk controls found")
            return None

    def get_all_risk_controls(self, taxonomy=None):
        """Get all risk control definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[RiskControl]
                Result containing a list of RiskControls
        """
        risk_control_instances = self._riskcontrols

        if taxonomy is not None:
            risk_control_instances = list(
                filter(
                    lambda risk_control: risk_control.isDefinedByTaxonomy == taxonomy,
                    risk_control_instances,
                )
            )

        return risk_control_instances

    def get_risk_control(self, id):
        """Get risk control definition from the LinkML by ID

        Args:
            id: str
                The string id for a risk control

        Returns:
            RiskControl
                Result containing a RiskControl
        """
        matching_risk_controls = list(
            filter(lambda risk_control: risk_control.id == id, self._riskcontrols)
        )

        if len(matching_risk_controls) > 0:
            return matching_risk_controls[0]
        else:
            print("No matching risk control found")
            return None

    def get_risk_incidents(self, taxonomy=None):
        """Get risk incident instances

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[RiskIncident]
                Result containing a list of RiskIncidents
        """
        risk_incident_instances = self._riskincidents or []

        if taxonomy is not None:
            risk_incident_instances = list(
                filter(
                    lambda risk_incident: risk_incident.isDefinedByTaxonomy == taxonomy,
                    risk_incident_instances,
                )
            )

        return risk_incident_instances

    def get_risk_incident(self, id):
        """Get risk incident instance by ID

        Args:
            id: str
                The string id for a risk incident

        Returns:
            RiskIncident
                Result containing a RiskIncident
        """
        risk_incident_instances = self._riskincidents or []

        matching_risk_incidents = list(
            filter(lambda risk_control: risk_control.id == id, risk_incident_instances)
        )

        if len(matching_risk_incidents) > 0:
            return matching_risk_incidents[0]
        else:
            print("No matching risk incident found")
            return None

    def get_related_risk_incidents(self, risk=None, risk_id=None, taxonomy=None):
        """Get related risk incidents for a risk

        Args:
            risk: (Optional) Risk
                The Risk object to find related incidents for
            risk_id: (Optional) str
            taxonomy: str
                (Optional) The string label for a taxonomy, to filter action results by

        Returns:
            list[RiskIncident]
                Result containing a list of the risk incidents which are marked as related to the specified AI risk
        """
        matching_risks = self._risks

        if risk is not None:
            matching_risks = [risk]
        if risk_id is not None:
            matching_risks = list(
                filter(lambda risk: risk.id == risk_id, matching_risks)
            )

        if len(matching_risks) > 0:
            risk: Risk = matching_risks[0]

            risk_incident_instances = self._riskincidents or []

            risk_incident_instances = list(
                filter(
                    lambda risk_incident: risk.id in risk_incident.refersToRisk,
                    risk_incident_instances,
                )
            )
            if taxonomy is not None:
                risk_incident_instances = list(
                    filter(
                        lambda risk_incident: risk_incident.isDefinedByTaxonomy
                        == taxonomy,
                        risk_incident_instances,
                    )
                )

            return risk_incident_instances
        else:
            print("No matching risk incidents found")
            return None

    def get_all_evaluations(self, taxonomy=None):
        """Get all evaluation definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[AiEval]
                Result containing a list of AiEval
        """
        evaluation_instances = self._evaluations

        if taxonomy is not None:
            evaluation_instances = list(
                filter(
                    lambda evaluation: evaluation.isDefinedByTaxonomy == taxonomy,
                    evaluation_instances,
                )
            )

        return evaluation_instances

    def get_evaluation(self, id):
        """Get evaluation definition from the LinkML by ID

        Args:
            id: str
                The string id for an evaluation

        Returns:
            AiEval
                Result containing an AiEval
        """
        matching_evaluations = list(
            filter(lambda evaluation: evaluation.id == id, self._evaluations)
        )

        if len(matching_evaluations) > 0:
            return matching_evaluations[0]
        else:
            print("No matching evaluation found")
            return None

    def get_related_evaluations(self, risk=None, risk_id=None, taxonomy=None):
        """Get related evaluations for a risk

        Args:
            risk: (Optional) Risk
                The Risk object to find related evaluations for
            risk_id: (Optional) str
            taxonomy: str
                (Optional) The string label for a taxonomy, to filter action results by

        Returns:
            list[AiEval]
                Result containing a list of the evaluations which are marked as related to the specified AI risk
        """
        matching_risks = self._risks

        if risk is not None:
            matching_risks = [risk]
        if risk_id is not None:
            matching_risks = list(
                filter(lambda risk: risk.id == risk_id, matching_risks)
            )

        if len(matching_risks) > 0:
            risk: Risk = matching_risks[0]

            evaluation_instances = self._evaluations or []

            evaluation_instances = list(
                filter(
                    lambda evaluation: risk.id in (evaluation.hasRelatedRisk or []),
                    evaluation_instances,
                )
            )
            if taxonomy is not None:
                evaluation_instances = list(
                    filter(
                        lambda evaluation: evaluation.isDefinedByTaxonomy == taxonomy,
                        evaluation_instances,
                    )
                )

            return evaluation_instances
        else:
            print("No matching evaluations found")
            return []

    def get_all_benchmark_metadata_cards(self, taxonomy=None, aieval_id=None):
        """Get all benchmark metadata definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy
            aieval_id: str
                (Optional) The string id for an AiEval that the benchmark describes

        Returns:
            list[AiEval]
                Result containing a list of BenchmarkMetadataCards
        """
        benchmark_metadata_card_instances = self._benchmarkmetadatacards or []

        if taxonomy is not None:
            benchmark_metadata_card_instances = list(
                filter(
                    lambda benchmark_metadata_card: benchmark_metadata_card.isDefinedByTaxonomy
                    == taxonomy,
                    benchmark_metadata_card_instances,
                )
            )

        if aieval_id is not None:
            benchmark_metadata_card_instances = list(
                filter(
                    lambda benchmark_metadata_card: benchmark_metadata_card.describesAiEval
                    == aieval_id,
                    benchmark_metadata_card_instances,
                )
            )

        return benchmark_metadata_card_instances

    def get_benchmark_metadata_card(self, id):
        """Get benchmark_metadata_card definition from the LinkML by ID

        Args:
            id: str
                The string id for a benchmark_metadata_card

        Returns:
            BenchmarkMetadataCard
                Result containing an BenchmarkMetadataCard
        """
        matching_benchmark_metadata_cards = list(
            filter(
                lambda benchmark_metadata_card: benchmark_metadata_card.id == id,
                self._benchmarkmetadatacards,
            )
        )

        if len(matching_benchmark_metadata_cards) > 0:
            return matching_benchmark_metadata_cards[0]
        else:
            print("No matching benchmark_metadata_card found")
            return None

    def get_documents(self, taxonomy=None):
        """Get all documentation definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Documentation]
                Result containing a list of Documentation entries
        """
        documentation_instances = self._documents or []

        if taxonomy is not None:
            documentation_instances = list(
                filter(
                    lambda document: document.isDefinedByTaxonomy == taxonomy,
                    documentation_instances,
                )
            )

        return documentation_instances

    def get_document(self, id):
        """Get document definition from the LinkML by ID

        Args:
            id: str
                The string id for a documentation entry

        Returns:
            Documentation
                Result containing a Documentation
        """
        matching_documentation_instances = list(
            filter(lambda document: document.id == id, self._documents)
        )

        if len(matching_documentation_instances) > 0:
            return matching_documentation_instances[0]
        else:
            print("No matching documents found")
            return None

    def get_datasets(self, taxonomy=None):
        """Get all dataset definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Dataset]
                Result containing a list of Dataset entries
        """
        dataset_instances = self._datasets or []

        if taxonomy is not None:
            dataset_instances = list(
                filter(
                    lambda dataset: dataset.isDefinedByTaxonomy == taxonomy,
                    dataset_instances,
                )
            )

        return dataset_instances

    def get_dataset(self, id):
        """Get dataset definition from the LinkML by ID

        Args:
            id: str
                The string id for a dataset entry

        Returns:
            Dataset
                Result containing a Dataset
        """
        matching_dataset_instances = list(
            filter(lambda dataset: dataset.id == id, self._datasets)
        )

        if len(matching_dataset_instances) > 0:
            return matching_dataset_instances[0]
        else:
            print("No matching dataset found")
            return []

    def get_stakeholders(self, taxonomy=None):
        """Get all stakeholder definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Stakeholder]
                Result containing a list of Stakeholder entries
        """
        stakeholder_instances = self._stakeholders or []

        if taxonomy is not None:
            stakeholder_instances = list(
                filter(
                    lambda stakeholder: stakeholder.isDefinedByTaxonomy == taxonomy,
                    stakeholder_instances,
                )
            )

        return stakeholder_instances

    def get_stakeholder(self, id):
        """Get Stakeholder definition from the LinkML by ID

        Args:
            id: str
                The string id for a Stakeholder entry

        Returns:
            Stakeholder
                Result containing a Stakeholder
        """
        matching_stakeholder_instances = list(
            filter(lambda stakeholder: stakeholder.id == id, self._stakeholders)
        )

        if len(matching_stakeholder_instances) > 0:
            return matching_stakeholder_instances[0]
        else:
            print("No matching stakeholder found")
            return []


    def get_llmintrinsics(self, taxonomy=None):
        """Get all LLMIntrinsic definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[LLMIntrinsic]
                Result containing a list of LLMIntrinsic entries
        """
        llmintrinsic_instances = self._llmintrinsics or []

        if taxonomy is not None:
            llmintrinsic_instances = list(
                filter(
                    lambda llmintrinsic: llmintrinsic.isDefinedByTaxonomy == taxonomy,
                    llmintrinsic_instances,
                )
            )

        return llmintrinsic_instances

    def get_llmintrinsic(self, id):
        """Get LLMIntrinsic definition from the LinkML by ID

        Args:
            id: str
                The string id for a LLMIntrinsic entry

        Returns:
            LLMIntrinsic
                Result containing a LLMIntrinsic
        """
        matching_llmintrinsic_instances = list(
            filter(lambda llmintrinsic: llmintrinsic.id == id, self._llmintrinsics)
        )

        if len(matching_llmintrinsic_instances) > 0:
            return matching_llmintrinsic_instances[0]
        else:
            print("No matching LLMIntrinsic found")
            return []

    def get_related_llmintrinsics(
        self, risk=None, tag=None, risk_id=None, name=None, taxonomy=None
    ):
        """Get llmintrinsics for a risk from the LinkML

        Args:
            risk: (Optional) Risk
                The Risk object to find related llmintrinsics for
            risk_id: (Optional) str
                The string ID identifying the risk to find related llmintrinsics for
            tag: (Optional) str
                The string tag identifying the risk to find related llmintrinsics for
            name: (Optional) str
                The string name identifying the risk to find related llmintrinsics for
            taxonomy: str
                (Optional) Only return llmintrinsics which come from this taxonomy

        Returns:
            list[LLMIntrinsics]
                Result containing a list of the llmintrinsics which are marked as related to the specified AI risk
        """
        matching_risks = self._risks
        llmintrinsic_instances = []

        if risk is not None:
            matching_risks = [risk]
        if tag is not None:
            matching_risks = list(filter(lambda risk: risk.tag == tag, matching_risks))
        if risk_id is not None:
            matching_risks = list(filter(lambda risk: risk.id == risk_id, matching_risks))
        if name is not None:
            matching_risks = list(
                filter(lambda risk: risk.name == name, matching_risks)
            )

        if len(matching_risks) > 0:
            risk: Risk = matching_risks[0]
            llmintrinsics = [x for x in self._llmintrinsics if (x.hasRelatedRisk is not None and risk.id in x.hasRelatedRisk)]
            if taxonomy is not None:
                llmintrinsic_instances = list(
                    filter(
                        lambda llmintrinsic: llmintrinsic.isDefinedByTaxonomy == taxonomy,
                        llmintrinsics,
                    )
                )


            return llmintrinsic_instances
        else:
            print("No matching intrinsics found")
            return []

    def get_adapters(self, taxonomy=None):
        """Get all Adapter definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[Adapter]
                Result containing a list of Adapter entries
        """
        adapter_instances = self._adapters or []

        if taxonomy is not None:
            adapter_instances = list(
                filter(
                    lambda adapter: adapter.isDefinedByTaxonomy == taxonomy,
                    adapter_instances,
                )
            )

        return adapter_instances

    def get_adapter(self, id):
        """Get Adapter definition from the LinkML by ID

        Args:
            id: str
                The string id for a Adapter entry

        Returns:
            Adapter
                Result containing an Adapter
        """
        matching_adapter_instances = list(
            filter(lambda adapter: adapter.id == id, self._adapters)
        )

        if len(matching_adapter_instances) > 0:
            return matching_adapter_instances[0]
        else:
            print("No matching adapter found")
            return []

    def get_llm_question_policies(self, taxonomy=None):
        """Get all LLM Question Policy definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy

        Returns:
            list[LLMQuestionPolicy]
                Result containing a list of LLM Question Policy entries
        """
        llm_question_policy_instances = self._llmquestionpolicies or []

        if taxonomy is not None:
            llm_question_policy_instances = list(
                filter(
                    lambda llmqp: llmqp.isDefinedByTaxonomy == taxonomy,
                    llm_question_policy_instances,
                )
            )

        return llm_question_policy_instances

    def get_llm_question_policy(self, id):
        """Get LLM Question Policy definition from the LinkML by ID

        Args:
            id: str
                The string id for a LLM Question Policy entry

        Returns:
            policy
                Result containing a LLM Question Policy
        """
        matching_llm_question_policy_instances = list(
            filter(lambda llm_question_policy: llm_question_policy.id == id, self._llmquestionpolicies)
        )

        if len(matching_llm_question_policy_instances) > 0:
            return matching_llm_question_policy_instances[0]
        else:
            print("No matching policy found")
            return []

    def get_principles(self, taxonomy=None, document=None):
        """Get all Principle definitions from the LinkML

        Args:
            taxonomy: str
                (Optional) The string label for a taxonomy
            document: str
                (Optional) The string label for a document

        Returns:
            list[Principle]
                Result containing a list of Principle entries
        """
        principle_instances = self._principles or []

        if taxonomy is not None:
            principle_instances = list(
                filter(
                    lambda principle: principle.isDefinedByTaxonomy == taxonomy,
                    principle_instances,
                )
            )
        if document is not None:
            principle_instances = list(
                filter(
                    lambda principle: principle.hasDocumentation == document,
                    principle_instances,
                )
            )

        return principle_instances

    def get_principle(self, id):
        """Get Principle definition from the LinkML by ID

        Args:
            id: str
                The string id for a Principle entry

        Returns:
            Principle
                Result containing a Principle
        """
        matching_principle_instances = list(
            filter(lambda principle: principle.id == id, self._principles)
        )

        if len(matching_principle_instances) > 0:
            return matching_principle_instances[0]
        else:
            print("No matching principle found")
            return []



from typing import Union

from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.utils.context_utils import CONTEXTS_PARAM_TYPE
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel


class LatexDumper(Dumper):

    def dump(
        self,
        element: Union[BaseModel, YAMLRoot],
        to_file: str,
        contexts: CONTEXTS_PARAM_TYPE = None,
        **kwargs,
    ) -> None:
        """
        Write element as latex to to_file

        Args:
            element: Union[BaseModel, YAMLRoot]
                LinkML object to be output
            to_file: str
                file to write to
            contexts: Optional[Union[CONTEXT_TYPE, List[CONTEXT_TYPE]]]
                 a list of JSON-LD contexts, which can be one of:
                    * the name of a JSON-LD file
                    * the URI of a JSON-lD file
                    * JSON-LD text
                    * A JsonObj object that contains JSON-LD
                    * A dictionary that contains JSON-LD
        """
        if isinstance(element, BaseModel):
            element = element.dict()

        super().dump(element, to_file, contexts=contexts, **kwargs)

    def dumps(self, element: Union[BaseModel, YAMLRoot], **_) -> str:
        """
        Return element as latex string

        Args:
            element: Union[BaseModel, YAMLRoot],
                LinkML object to be emitted
            _:
                method specific arguments

        Returns:
            str
        """
        if isinstance(element, BaseModel):
            element = element.dict()
            element = self._to_tex_from_dict(element)
        return element

    def _to_tex_from_dict(self, element_dict) -> str:
        """
        Format as tex

        Args:
            element_dict: dict

        Returns:
            str
        """
        tex = [
            "\\section{AI Risk Atlas Definitions}\n",
            "The below is a catalog of potential risks when working with generative AI, foundation models, and machine learning models.\n\n\n",
        ]

        for risk in element_dict["risks"]:
            tex.append("\\begin{definitionbox}{" + risk["name"] + "}\n")
            tex.append(
                risk["description"].replace("’", "'").replace(" ", " ")
                + "\\newline\\newline\n"
            )
            tex.append(
                "\\textbf{Concern: }"
                + risk["concern"].replace("’", "'").replace(" ", " ")
                + "\\newline\\newline\n"
            )
            tex.append(
                "\\textbf{Type: }"
                + risk["type"].replace("’", "'").replace(" ", " ")
                + "\\newline\n"
            )
            tex.append(
                "\\textbf{Descriptor: }" + risk["descriptor"] + " \\newline\\newline\n"
            )
            tex.append("\\textbf{Implementation details: }" + " \\newline\n")
            tex.append("ID: " + risk["id"] + " \\newline\n")
            tex.append("Tag: " + risk["tag"] + " \\newline\n")
            tex.append(
                "URI:  \\href{"
                + risk["url"]
                + "}{IBM AI Risk Atlas - "
                + risk["name"]
                + "}\\newline\n"
            )
            tex.append("\\end{definitionbox}\n")
        tex.append("\\end{document}\n")
        tex_all = "".join(tex)
        return tex_all



import datetime
import json
import re
from typing import List

from sssom_schema import EntityReference, Mapping
from txtai import Embeddings

from ai_atlas_nexus.ai_risk_ontology.datamodel.ai_risk_ontology import Risk
from ai_atlas_nexus.blocks.inference import InferenceEngine
from ai_atlas_nexus.blocks.inference.params import TextGenerationInferenceOutput
from ai_atlas_nexus.blocks.prompt_builder import ZeroShotPromptBuilder
from ai_atlas_nexus.blocks.prompt_response_schema import LIST_OF_STR_SCHEMA
from ai_atlas_nexus.blocks.prompt_templates import RISK_IDENTIFICATION_TEMPLATE
from ai_atlas_nexus.blocks.risk_detector import GenericRiskDetector
from ai_atlas_nexus.blocks.risk_mapping import RiskMappingBase
from ai_atlas_nexus.metadata_base import MappingMethod


class RiskMapper(RiskMappingBase):

    def _bucket_semantic_score(self, score: int):
        """A simplistic method to bucket the scores

        Args:
            score: int
                The semantic score to be processed

        Returns:
            str
                A representation of the relationship
        """
        relationship = "noMatch"
        if score == 100:
            relationship = "skos:exactMatch"
        elif (score <= 100) and (score > 80):
            relationship = "skos:closeMatch"
        elif (score <= 80) and (score > 20):
            relationship = "skos:relatedMatch"

        return relationship

    def _format_with_curie(self, curie_prefix, entity_id):
        """Format the string with curie prefix

        Args:
            curie_prefix: str
                The curie prefix
            entity_id: str
                The linkml instance id

        Returns:
            EntityReference
                A formatted string
        """
        s = curie_prefix.strip() + ":" + entity_id.strip()
        return EntityReference(s)

    def generate(
        self,
        new_risks: list[Risk],
        existing_risks: list[Risk],
        inference_engine: InferenceEngine,
        new_prefix: str,
        mapping_method: MappingMethod,
    ) -> list[Mapping]:
        """Generate a list of mappings between two lists of risks
        Args:
            new_risks: list[Risk]
                A new set of risks
            existing_risks: list[Risk],
                Secondary list, this should be the list of existing risks in RAN
            inference_engine: (Optional)Union[InferenceEngine | None]:
                An LLM inference engine to infer risks from the usecases.
            new_prefix: str
                A curie prefix for the new list
            mapping_method: MappingMethod
                The method to generate the mapping

        Returns:
            list[Mapping]
        """
        mappings = []

        data = []
        taxonomy_for_mapping = {}
        for risk in existing_risks:
            # this embedding is just using name and description, not any other attributes
            data.append(
                "ID: "
                + risk.id
                + ", Name: "
                + risk.name
                + ", Description: "
                + risk.description
            )
            taxonomy_for_mapping[risk.id] = risk.isDefinedByTaxonomy

        if mapping_method == MappingMethod.SEMANTIC:
            # create an embedding with existing risk data

            embeddings = Embeddings(path="sentence-transformers/nli-mpnet-base-v2")
            embeddings.index(data)

            # Run an embeddings search for each new risk
            for nr in new_risks:
                # Extract uid of first result
                # search result format: (uid, score)
                query = (
                    "ID: "
                    + nr.id
                    + ", Name: "
                    + nr.name
                    + ", Description: "
                    + nr.description
                )

                # embedding search returns list of (id, score) for index search, e.g. [(54, 0.6546236872673035), (48, 0.5914335250854492)]
                # let's just use the top match for now

                top_result = embeddings.search(query, 5)[0][0]  # id of top result
                s = data[top_result]  # string data belonging to that ID
                result_id = re.search("ID:(.*), Name", s)
                result_name = re.search("Name:(.*), Description:", s)

                mapping = Mapping(
                    subject_id=self._format_with_curie(nr.isDefinedByTaxonomy, nr.id),
                    subject_label=nr.name,
                    predicate_id=self._bucket_semantic_score(top_result),
                    object_id=self._format_with_curie(
                        taxonomy_for_mapping[result_id.group(1).strip()],
                        result_id.group(1),
                    ),
                    object_label=result_name.group(1),
                    mapping_justification="semapv:SemanticSimilarityThresholdMatching",
                    similarity_score=top_result,
                    mapping_date=datetime.date.today(),
                    author_id="AI_Atlas_Nexus_System",
                    comment="Autogenerated via semantic similarity script",
                )
                mappings.append(mapping)

        elif mapping_method == MappingMethod.INFERENCE:
            # this query is just using name and description, not any other attributes
            usecases = [
                (
                    "ID: "
                    + nr.id
                    + ", Name: "
                    + nr.name
                    + ", Description: "
                    + nr.description
                )
                for nr in new_risks
            ]

            risk_detector = GenericRiskDetector(
                risks=existing_risks,
                inference_engine=self.inference_engine,
                cot_examples=None,
            )

            rls = risk_detector.detect(usecases)

            for (
                index,
                rl,
            ) in enumerate(rls):
                for risk in rl:
                    mapping = Mapping(
                        subject_id=self._format_with_curie(
                            new_risks[index].isDefinedByTaxonomy, new_risks[index].id
                        ),
                        subject_label=new_risks[index].name,
                        predicate_id="skos:relatedMatch",  #  opted for this here as there is no way to assess relatedness of match with current template
                        object_id=self._format_with_curie(
                            taxonomy_for_mapping[risk.id.strip()], risk.id
                        ),
                        object_label=risk.name,
                        mapping_justification="semapv:LLMBasedMatching",
                        mapping_date=datetime.date.today(),
                        author_id="AI_Atlas_Nexus_System",
                        comment="Autogenerated via LLM based matching script",
                    )
                    mappings.append(mapping)

        return mappings



import os
from typing import Dict, List, Union

import httpx
from dotenv import load_dotenv
from openai import NotFoundError

from ai_atlas_nexus.blocks.inference.base import InferenceEngine
from ai_atlas_nexus.blocks.inference.params import (
    InferenceEngineCredentials,
    OpenAIChatCompletionMessageParam,
    RITSInferenceEngineParams,
    TextGenerationInferenceOutput,
)
from ai_atlas_nexus.blocks.inference.postprocessing import postprocess
from ai_atlas_nexus.metadata_base import InferenceEngineType
from ai_atlas_nexus.toolkit.job_utils import run_parallel


# load .env file to environment
load_dotenv()


class RITSInferenceEngine(InferenceEngine):

    _inference_engine_type = InferenceEngineType.RITS
    _inference_engine_parameter_class = RITSInferenceEngineParams

    def prepare_credentials(
        self, credentials: Union[Dict, InferenceEngineCredentials]
    ) -> InferenceEngineCredentials:
        api_key = credentials.get(
            "api_key", os.environ.get(f"{self._inference_engine_type}_API_KEY", None)
        )
        assert api_key, (
            f"Error while trying to run {self._inference_engine_type}. "
            f"Please set the env variable: '{self._inference_engine_type}_API_KEY' or pass api_key to credentials."
        )

        api_url = credentials.get(
            "api_url", os.environ.get(f"{self._inference_engine_type}_API_URL", None)
        )
        assert api_url, (
            f"Error while trying to run {self._inference_engine_type}. "
            f"Please set the env variable: '{self._inference_engine_type}_API_URL' or pass api_url to credentials."
        )

        return InferenceEngineCredentials(api_key=api_key, api_url=api_url)

    def create_client(self, credentials):
        from openai import OpenAI

        model_name_for_endpoint = (
            self.model_name_or_path.split("/")[-1].lower().replace(".", "-")
        )
        return OpenAI(
            api_key=credentials["api_key"],
            base_url=f"{credentials['api_url']}/{model_name_for_endpoint}/v1",
            default_headers={"RITS_API_KEY": credentials["api_key"]},
            timeout=httpx.Timeout(None, connect=1.0),
        )

    def ping(self):
        try:
            self.client.models.list()
        except NotFoundError:
            raise Exception(f"Model `{self.model_name_or_path}` not available.")

    @postprocess
    def generate(
        self,
        prompts: List[str],
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> List[TextGenerationInferenceOutput]:
        return self.chat(prompts, response_format, postprocessors, verbose)

    @postprocess
    def chat(
        self,
        messages: Union[
            List[OpenAIChatCompletionMessageParam],
            List[str],
        ],
        response_format=None,
        postprocessors=None,
        verbose=True,
    ) -> TextGenerationInferenceOutput:

        def chat_response(messages):
            response = self.client.chat.completions.create(
                messages=self._to_openai_format(messages),
                model=self.model_name_or_path,
                response_format=self._create_schema_format(response_format),
                **self.parameters,
            )
            return self._prepare_chat_output(response)

        return run_parallel(
            chat_response,
            messages,
            f"Inferring with {self._inference_engine_type}",
            self.concurrency_limit,
            verbose=verbose,
        )

    def _prepare_chat_output(self, response):
        return TextGenerationInferenceOutput(
            prediction=response.choices[0].message.content,
            input_tokens=response.usage.total_tokens,
            output_tokens=response.usage.completion_tokens,
            stop_reason=response.choices[0].finish_reason,
            model_name_or_path=self.model_name_or_path,
            logprobs=(
                {
                    output.token: output.logprob
                    for output in response.choices[0].logprobs.content
                }
                if response.choices[0].logprobs
                else None
            ),
            inference_engine=str(self._inference_engine_type),
        )

    def _create_schema_format(self, response_format):
        if response_format:
            return {
                "type": "json_schema",
                "json_schema": {
                    "name": "RITS_schema",
                    "schema": response_format,
                },
            }
        else:
            return None



# Standard
from enum import Enum, unique


@unique
class InferenceEngineType(str, Enum):
    """Enum to contain possible values for inference engine types"""

    RITS = "RITS"
    WML = "WML"
    VLLM = "VLLM"
    OLLAMA = "OLLAMA"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.name, cls))

    def __str__(self):
        return self.name


@unique
class MappingMethod(str, Enum):
    """Enum to contain possible values for risk mapping methods"""

    SEMANTIC = "SEMANTIC"
    INFERENCE = "INFERENCE"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.name, cls))

    def __str__(self):
        return self.name
